import re
import pathlib

from setuptools import setup


HERE = pathlib.Path(__file__).parent

txt = (HERE / 'danmu_abc' / '__init__.py').read_text('utf-8')
try:
    version = re.findall(r"^__version__ = '([^']+)'\r?$", txt, re.M)[0]
except IndexError:
    raise RuntimeError('Unable to determine the version.')


def read(f):
    return (HERE / f).read_text('utf-8').strip()


setup(
    name='danmu_abc',
    packages=['danmu_abc'],
    version=version,
    description='Danmu for humans.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    license='MIT',
    author='yjqang',
    author_email='yjqiang01@gmail.com',
    url='https://github.com/yjqiang/danmu',
    keywords=[
        'danmu', 'huya', 'bilibili', 'douyu'
    ],
    install_requires=[
        'aiohttp>=3.6.2,<4.0.0',
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    python_requires='>=3.6',
)



import json
import asyncio
from typing import Optional, Any
from urllib.parse import urlparse
from abc import ABC, abstractmethod

from aiohttp import ClientSession, WSMsgType, ClientError


class Conn(ABC):
    __slots__ = ('_receive_timeout',)

    # receive_timeout 推荐为 heartbeat 间隔加 10s 或 5s
    @abstractmethod
    def __init__(self, receive_timeout: Optional[float] = None):
        self._receive_timeout = receive_timeout

    @abstractmethod
    async def open(self) -> bool:
        return False

    @abstractmethod
    async def close(self) -> bool:
        return True
        
    # 用于永久 close 之后一些数据清理等
    @abstractmethod
    async def clean(self) -> None:
        pass

    @abstractmethod
    async def send_bytes(self, bytes_data: bytes) -> bool:
        return True

    @abstractmethod
    async def read_bytes(self) -> Optional[bytes]:
        return None

    @abstractmethod
    async def read_json(self) -> Any:
        return None
        
        
class TcpConn(Conn):
    __slots__ = ('_host', '_port', '_reader', '_writer')

    # url 格式 tcp://hostname:port
    def __init__(self, url: str, receive_timeout: Optional[float] = None):
        super().__init__(receive_timeout)
        result = urlparse(url)
        if result.scheme != 'tcp':
            raise TypeError(f'url scheme must be tcp ({result.scheme})')

        self._host = result.hostname
        self._port = result.port
        self._reader = None
        self._writer = None
        
    async def open(self) -> bool:
        try:
            self._reader, self._writer = await asyncio.wait_for(
                asyncio.open_connection(self._host, self._port), timeout=3)
        except (OSError, asyncio.TimeoutError):
            return False
        return True
        
    async def close(self) -> bool:
        if self._writer is not None:
            self._writer.close()
            # py3.7 才有（妈的你们真的磨叽）
            # await self._writer.wait_closed()
        return True
    
    async def clean(self) -> None:
        pass
        
    async def send_bytes(self, bytes_data: bytes) -> bool:
        try:
            self._writer.write(bytes_data)
            await self._writer.drain()
        except OSError:
            return False
        except asyncio.CancelledError:
            # print('asyncio.CancelledError', 'send_bytes')
            return False
        return True
        
    async def read_bytes(
            self,
            n: Optional[int] = None) -> Optional[bytes]:
        # 不支持的原因是，tcp 会切数据，自己拼装过于复杂
        if n is None or n <= 0:
            return None
        try:
            bytes_data = await asyncio.wait_for(
                self._reader.readexactly(n), timeout=self._receive_timeout)
        except (OSError, asyncio.TimeoutError, asyncio.streams.IncompleteReadError):
            return None
        except asyncio.CancelledError:
            # print('asyncio.CancelledError', 'read_bytes')
            return None
        return bytes_data
        
    async def read_json(
            self,
            n: Optional[int] = None) -> Any:
        data = await self.read_bytes(n)
        if not data:
            return None
        return json.loads(data.decode('utf8'))
                

class WsConn(Conn):
    __slots__ = ('_url', '_is_sharing_session', '_session', '_ws_receive_timeout', '_ws_heartbeat', '_ws')

    # url 格式 ws://hostname:port/… 或者 wss://hostname:port/…
    def __init__(
            self, url: str,
            receive_timeout: Optional[float] = None,
            session: Optional[ClientSession] = None,
            ws_receive_timeout: Optional[float] = None,  # 自动 ping pong 时候用的
            ws_heartbeat: Optional[float] = None):  # 自动 ping pong 时候用的
        super().__init__(receive_timeout)
        result = urlparse(url)
        if result.scheme != 'ws' and result.scheme != 'wss':
            raise TypeError(f'url scheme must be websocket ({result.scheme})')
        self._url = url
        
        if session is None:
            self._is_sharing_session = False
            self._session = ClientSession()
        else:
            self._is_sharing_session = True
            self._session = session
        self._ws_receive_timeout = ws_receive_timeout
        self._ws_heartbeat = ws_heartbeat
        self._ws = None

    async def open(self) -> bool:
        try:
            self._ws = await asyncio.wait_for(
                self._session.ws_connect(
                    self._url,
                    receive_timeout=self._ws_receive_timeout,
                    heartbeat=self._ws_heartbeat), timeout=3)
        except (ClientError, asyncio.TimeoutError):
            return False
        return True
        
    async def close(self) -> bool:
        if self._ws is not None:
            await self._ws.close()
        return True
        
    async def clean(self) -> None:
        if not self._is_sharing_session:
            await self._session.close()
        
    async def send_bytes(self, bytes_data: bytes) -> bool:
        try:
            await self._ws.send_bytes(bytes_data)
        except ClientError:
            return False
        except asyncio.CancelledError:
            return False
        return True
                
    async def read_bytes(self) -> Optional[bytes]:
        try:
            msg = await asyncio.wait_for(
                self._ws.receive(), timeout=self._receive_timeout)
            if msg.type == WSMsgType.BINARY:
                return msg.data
        except (ClientError, asyncio.TimeoutError):
            return None
        except asyncio.CancelledError:
            # print('asyncio.CancelledError', 'read_bytes')
            return None
        return None

    async def read_json(self) -> Any:
        try:
            msg = await asyncio.wait_for(
                self._ws.receive(), timeout=self._receive_timeout)
            if msg.type == WSMsgType.TEXT:
                return json.loads(msg.data)
            elif msg.type == WSMsgType.BINARY:
                return json.loads(msg.data.decode('utf8'))
        except (ClientError, asyncio.TimeoutError):
            return None
        except asyncio.CancelledError:
            # print('asyncio.CancelledError', 'read_json')
            return None
        return None



import asyncio
from typing import Optional, Callable
from abc import ABC, abstractmethod

from .conn import Conn


class Client(ABC):
    __slots__ = (
        '_loop', '_area_id', '_conn', '_opening_lock', '_waiting_end', '_waiting_pause', '_closed',
        '_heartbeat', '_task_main', '_funcs_task', '_logger_info'
    )

    def __init__(
            self, area_id: int, conn: Conn, heartbeat: float = 30.0, loop: asyncio.AbstractEventLoop = None,
            logger_info: Callable = print
    ):
        self._loop = asyncio.get_event_loop() if loop is None else loop
        self._logger_info = logger_info
         
        self._area_id = area_id
        self._conn = conn
            
        # 建立连接过程中难以处理重设置房间或断线等问题
        self._opening_lock = asyncio.Lock()
        self._waiting_end: Optional[asyncio.Future] = None
        self._waiting_pause: Optional[asyncio.Future] = None
        self._closed = False

        self._heartbeat = heartbeat  # 心跳间隔

        self._task_main = None
        # 除了 _job_main 和 _job_heartbeat
        self._funcs_task = []

    # 建立连接并且 _say_hello
    async def _job_open(self) -> bool:
        if await self._conn.open():
            return await self._one_hello()
        return False

    # 建立连接后进行一次 auth
    @abstractmethod
    async def _one_hello(self) -> bool:
        pass

    # 持续心跳
    async def _job_heartbeat(self) -> None:
        try:
            while True:
                if not await self._one_heartbeat():
                    return
                await asyncio.sleep(self._heartbeat)
        except asyncio.CancelledError:
            return

    # 发送一次心跳包
    @abstractmethod
    async def _one_heartbeat(self) -> bool:
        pass

    # 循环读取
    async def _job_main(self) -> None:
        while self._waiting_pause is None:
            if not await self._one_read():
                return

    # 读一次数据（完整数据包括头尾）
    @abstractmethod
    async def _one_read(self) -> bool:
        pass

    # 关闭当前连接，并非永久关闭！不包含 clean 工作等！
    async def _job_close(self) -> None:
        await self._conn.close()

    @staticmethod
    async def _prepare_client() -> bool:
        return True

    async def run_forever(self) -> None:
        self._waiting_end = self._loop.create_future()
        while not self._closed:
            self._logger_info(f'正在启动 {self._area_id} 号数据连接')
            if self._waiting_pause is not None:
                self._logger_info(f'暂停启动 {self._area_id} 号数据连接，等待 RESUME 指令')
                await self._waiting_pause
            
            async with self._opening_lock:
                if self._closed:
                    self._logger_info(f'{self._area_id} 号数据连接确认收到关闭信号，正在处理')
                    break
                # 未成功建立数据连接，循环重试
                if await self._prepare_client() and await self._job_open():
                    tasks = [self._loop.create_task(i()) for i in self._funcs_task]

                    self._task_main = self._loop.create_task(self._job_main())
                    tasks.append(self._task_main)

                    task_heartbeat = self._loop.create_task(self._job_heartbeat())
                    tasks.append(task_heartbeat)
                else:
                    continue
                
            _, pending = await asyncio.wait(
                tasks, return_when=asyncio.FIRST_COMPLETED)
            self._logger_info(f'{self._area_id} 号数据连接异常或主动断开，正在处理剩余信息')
            for i in pending:
                if i != self._task_main:
                    i.cancel()
            await self._job_close()
            if pending:
                await asyncio.wait(pending)
            self._logger_info(f'{self._area_id} 号数据连接退出，剩余任务处理完毕')
        await self._conn.clean()
        self._waiting_end.set_result(True)

    @property
    def paused(self) -> bool:
        return self._waiting_pause is not None

    # 非主动 cancel，而是静静等下一个主循环
    def pause(self) -> None:
        if self._waiting_pause is None:
            self._waiting_pause = self._loop.create_future()
            
    def resume(self) -> None:
        if self._waiting_pause is not None:
            self._waiting_pause.set_result(True)
            self._waiting_pause = None
            
    async def close_and_clean(self) -> bool:
        if not self._closed:
            self._closed = True
            async with self._opening_lock:
                await self._job_close()

            if self._waiting_end is not None:
                await self._waiting_end
            return True
        return False
